"use strict";
/// <reference path="./tauProlog.d.ts"/>
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const fslib_1 = require("@yarnpkg/fslib");
const tau_prolog_1 = __importDefault(require("tau-prolog"));
const tauModule_1 = require("./tauModule");
var DependencyType;
(function (DependencyType) {
    DependencyType["Dependencies"] = "dependencies";
    DependencyType["DevDependencies"] = "devDependencies";
    DependencyType["PeerDependencies"] = "peerDependencies";
})(DependencyType = exports.DependencyType || (exports.DependencyType = {}));
const DEPENDENCY_TYPES = [
    DependencyType.Dependencies,
    DependencyType.DevDependencies,
    DependencyType.PeerDependencies,
];
function extractErrorImpl(value) {
    if (value instanceof tau_prolog_1.default.type.Num)
        return value.value;
    if (value instanceof tau_prolog_1.default.type.Term) {
        if (value.args.length === 0)
            return value.id;
        switch (value.indicator) {
            case `throw/1`:
                return extractErrorImpl(value.args[0]);
            case `error/1`:
                return extractErrorImpl(value.args[0]);
            case `error/2`:
                return Object.assign(extractErrorImpl(value.args[0]), ...extractErrorImpl(value.args[1]));
            case `syntax_error/1`:
                return new core_1.ReportError(core_1.MessageName.PROLOG_SYNTAX_ERROR, `Syntax error: ${extractErrorImpl(value.args[0])}`);
            case `existence_error/2`:
                return new core_1.ReportError(core_1.MessageName.PROLOG_EXISTENCE_ERROR, `Existence error: ${extractErrorImpl(value.args[0])} ${extractErrorImpl(value.args[1])} not found`);
            case `line/1`:
                return { line: extractErrorImpl(value.args[0]) };
            case `column/1`:
                return { column: extractErrorImpl(value.args[0]) };
            case `found/1`:
                return { found: extractErrorImpl(value.args[0]) };
            case `./2`:
                return [extractErrorImpl(value.args[0])].concat(extractErrorImpl(value.args[1]));
            case `//2`:
                return `${extractErrorImpl(value.args[0])}/${extractErrorImpl(value.args[1])}`;
        }
    }
    throw `couldn't pretty print because of unsupported node ${value}`;
}
function extractError(val) {
    let err;
    try {
        err = extractErrorImpl(val);
    }
    catch (caught) {
        if (typeof caught === `string`) {
            throw new core_1.ReportError(core_1.MessageName.PROLOG_UNKNOWN_ERROR, `Unknown error: ${val} (note: ${caught})`);
        }
        else {
            throw caught;
        }
    }
    if (typeof err.line !== `undefined` && typeof err.column !== `undefined`)
        err.message += ` at line ${err.line}, column ${err.column}`;
    return err;
}
// Node 8 doesn't have Symbol.asyncIterator
// https://github.com/Microsoft/TypeScript/issues/14151#issuecomment-280812617
if (Symbol.asyncIterator == null)
    Symbol.asyncIterator = Symbol.for('Symbol.asyncIterator');
class Session {
    constructor(project, source) {
        this.session = tau_prolog_1.default.create();
        tauModule_1.linkProjectToSession(this.session, project);
        this.session.consult(source);
    }
    fetchNextAnswer() {
        return new Promise(resolve => {
            this.session.answer((result) => {
                resolve(result);
            });
        });
    }
    makeQuery(query) {
        return __asyncGenerator(this, arguments, function* makeQuery_1() {
            const parsed = this.session.query(query);
            if (parsed !== true)
                throw extractError(parsed);
            while (true) {
                const answer = yield __await(this.fetchNextAnswer());
                if (!answer)
                    break;
                if (answer.id === `throw`)
                    throw extractError(answer);
                yield yield __await(answer);
            }
        });
    }
}
function parseLink(link) {
    if (link.id === `null`) {
        return null;
    }
    else {
        return `${link.toJavaScript()}`;
    }
}
function parseLinkToJson(link) {
    if (link.id === `null`) {
        return null;
    }
    else {
        const val = link.toJavaScript();
        if (typeof val !== `string`)
            return JSON.stringify(val);
        try {
            return JSON.stringify(JSON.parse(val));
        }
        catch (_a) {
            return JSON.stringify(val);
        }
    }
}
class Constraints {
    constructor(project) {
        this.source = ``;
        this.project = project;
        const constraintsPath = project.configuration.get(`constraintsPath`);
        if (fslib_1.xfs.existsSync(constraintsPath)) {
            this.source = fslib_1.xfs.readFileSync(constraintsPath, `utf8`);
        }
    }
    static async find(project) {
        return new Constraints(project);
    }
    getProjectDatabase() {
        let database = ``;
        for (const dependencyType of DEPENDENCY_TYPES)
            database += `dependency_type(${dependencyType}).\n`;
        for (const workspace of this.project.workspacesByCwd.values()) {
            const relativeCwd = workspace.relativeCwd;
            database += `workspace(${escape(relativeCwd)}).\n`;
            database += `workspace_ident(${escape(relativeCwd)}, ${escape(core_2.structUtils.stringifyIdent(workspace.locator))}).\n`;
            database += `workspace_version(${escape(relativeCwd)}, ${escape(workspace.manifest.version)}).\n`;
            for (const dependencyType of DEPENDENCY_TYPES) {
                for (const dependency of workspace.manifest[dependencyType].values()) {
                    database += `workspace_has_dependency(${escape(relativeCwd)}, ${escape(core_2.structUtils.stringifyIdent(dependency))}, ${escape(dependency.range)}, ${dependencyType}).\n`;
                }
            }
        }
        // Add default never matching predicates to prevent prolog instantiation errors
        // when constraints run in an empty workspace
        database += `workspace(_) :- false.\n`;
        database += `workspace_ident(_, _) :- false.\n`;
        database += `workspace_version(_, _) :- false.\n`;
        // Add a default never matching predicate to prevent prolog instantiation errors
        // when constraints run in a workspace without dependencies
        database += `workspace_has_dependency(_, _, _, _) :- false.\n`;
        return database;
    }
    getDeclarations() {
        let declarations = ``;
        // (Cwd, DependencyIdent, DependencyRange, DependencyType)
        declarations += `gen_enforced_dependency(_, _, _, _) :- false.\n`;
        // (Cwd, Path, Value)
        declarations += `gen_enforced_field(_, _, _) :- false.\n`;
        return declarations;
    }
    get fullSource() {
        return `${this.getProjectDatabase()}\n${this.source}\n${this.getDeclarations()}`;
    }
    createSession() {
        return new Session(this.project, this.fullSource);
    }
    async process() {
        const session = this.createSession();
        return {
            enforcedDependencies: await this.genEnforcedDependencies(session),
            enforcedFields: await this.genEnforcedFields(session),
        };
    }
    async genEnforcedDependencies(session) {
        var e_1, _a;
        let enforcedDependencies = [];
        try {
            for (var _b = __asyncValues(session.makeQuery(`workspace(WorkspaceCwd), dependency_type(DependencyType), gen_enforced_dependency(WorkspaceCwd, DependencyIdent, DependencyRange, DependencyType).`)), _c; _c = await _b.next(), !_c.done;) {
                const answer = _c.value;
                const workspaceCwd = fslib_1.ppath.resolve(this.project.cwd, parseLink(answer.links.WorkspaceCwd));
                const dependencyRawIdent = parseLink(answer.links.DependencyIdent);
                const dependencyRange = parseLink(answer.links.DependencyRange);
                const dependencyType = parseLink(answer.links.DependencyType);
                if (workspaceCwd === null || dependencyRawIdent === null)
                    throw new Error(`Invalid rule`);
                const workspace = this.project.getWorkspaceByCwd(workspaceCwd);
                const dependencyIdent = core_2.structUtils.parseIdent(dependencyRawIdent);
                enforcedDependencies.push({ workspace, dependencyIdent, dependencyRange, dependencyType });
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) await _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return core_2.miscUtils.sortMap(enforcedDependencies, [
            ({ dependencyRange }) => dependencyRange !== null ? `0` : `1`,
            ({ workspace }) => core_2.structUtils.stringifyIdent(workspace.locator),
            ({ dependencyIdent }) => core_2.structUtils.stringifyIdent(dependencyIdent),
        ]);
    }
    async genEnforcedFields(session) {
        var e_2, _a;
        let enforcedFields = [];
        try {
            for (var _b = __asyncValues(session.makeQuery(`workspace(WorkspaceCwd), gen_enforced_field(WorkspaceCwd, FieldPath, FieldValue).`)), _c; _c = await _b.next(), !_c.done;) {
                const answer = _c.value;
                const workspaceCwd = fslib_1.ppath.resolve(this.project.cwd, parseLink(answer.links.WorkspaceCwd));
                const fieldPath = parseLink(answer.links.FieldPath);
                let fieldValue = parseLinkToJson(answer.links.FieldValue);
                if (workspaceCwd === null || fieldPath === null)
                    throw new Error(`Invalid rule`);
                const workspace = this.project.getWorkspaceByCwd(workspaceCwd);
                enforcedFields.push({ workspace, fieldPath, fieldValue });
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) await _a.call(_b);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return core_2.miscUtils.sortMap(enforcedFields, [
            ({ workspace }) => core_2.structUtils.stringifyIdent(workspace.locator),
            ({ fieldPath }) => fieldPath,
        ]);
    }
    query(query) {
        return __asyncGenerator(this, arguments, function* query_1() {
            var e_3, _a;
            const session = this.createSession();
            try {
                for (var _b = __asyncValues(session.makeQuery(query)), _c; _c = yield __await(_b.next()), !_c.done;) {
                    const answer = _c.value;
                    const parsedLinks = {};
                    for (const [variable, value] of Object.entries(answer.links)) {
                        if (variable !== `_`) {
                            parsedLinks[variable] = parseLink(value);
                        }
                    }
                    yield yield __await(parsedLinks);
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) yield __await(_a.call(_b));
                }
                finally { if (e_3) throw e_3.error; }
            }
        });
    }
}
exports.Constraints = Constraints;
function escape(what) {
    if (typeof what === `string`) {
        return `'${what}'`;
    }
    else {
        return `[]`;
    }
}
