"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const Constraints_1 = require("../../Constraints");
// eslint-disable-next-line arca/no-default-export
class ConstraintsQueryCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.json = false;
    }
    async execute() {
        const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
        const { project } = await core_1.Project.find(configuration, this.context.cwd);
        const constraints = await Constraints_1.Constraints.find(project);
        let query = this.query;
        if (!query.endsWith(`.`))
            query = `${query}.`;
        const report = await core_2.StreamReport.start({
            configuration,
            json: this.json,
            stdout: this.context.stdout,
        }, async (report) => {
            var e_1, _a;
            try {
                for (var _b = __asyncValues(constraints.query(query)), _c; _c = await _b.next(), !_c.done;) {
                    const result = _c.value;
                    const lines = Array.from(Object.entries(result));
                    const lineCount = lines.length;
                    const maxVariableNameLength = lines.reduce((max, [variableName]) => Math.max(max, variableName.length), 0);
                    for (let i = 0; i < lineCount; i++) {
                        const [variableName, value] = lines[i];
                        report.reportInfo(null, `${getLinePrefix(i, lineCount)}${variableName.padEnd(maxVariableNameLength, ` `)} = ${valueToString(value)}`);
                    }
                    report.reportJson(result);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) await _a.call(_b);
                }
                finally { if (e_1) throw e_1.error; }
            }
        });
        return report.exitCode();
    }
}
ConstraintsQueryCommand.usage = clipanion_1.Command.Usage({
    category: `Constraints-related commands`,
    description: `query the constraints fact database`,
    details: `
      This command will output all matches to the given prolog query.

      If the \`--json\` flag is set the output will follow a JSON-stream output also known as NDJSON (https://github.com/ndjson/ndjson-spec).
    `,
    examples: [[
            `List all dependencies throughout the workspace`,
            `yarn constraints query 'workspace_has_dependency(_, DependencyName, _, _).'`,
        ]],
});
__decorate([
    clipanion_1.Command.Boolean(`--json`)
], ConstraintsQueryCommand.prototype, "json", void 0);
__decorate([
    clipanion_1.Command.String()
], ConstraintsQueryCommand.prototype, "query", void 0);
__decorate([
    clipanion_1.Command.Path(`constraints`, `query`)
], ConstraintsQueryCommand.prototype, "execute", null);
exports.default = ConstraintsQueryCommand;
function valueToString(value) {
    if (typeof value !== `string`)
        return `${value}`;
    if (value.match(/^[a-zA-Z][a-zA-Z0-9_]+$/))
        return value;
    return `'${value}'`;
}
function getLinePrefix(index, count) {
    const isFirst = index === 0;
    const isLast = index === count - 1;
    if (isFirst && isLast)
        return ``;
    if (isFirst)
        return `┌ `;
    if (isLast)
        return `└ `;
    return `│ `;
}
